#!/usr/bin/env python

import re
import requests

from urllib3.util import parse_url
from bs4          import BeautifulSoup
from base64       import b64decode, b64encode
from stat         import S_ISREG

from os import (
    curdir,
    getenv,
    makedirs,
    sep,
    stat,
    unlink,
    rename
)

from .exceptions import (
    OutOfBoundError,
    UnsupportedEngine,
    CannotGoBack
)

from .      import utils
from .image import Image

class Search:

    # Static Vars
    encoding = utils.preferred_encoding()

    search_urls = {
        'yahoo': 'https://search.yahoo.com/search/?p={0}',
        'duckduckgo': 'https://html.duckduckgo.com/html/?q={0}',
        'google': 'https://www.google.com/search?q={0}',
    }

    @staticmethod
    def check_engine(engine):
        if not engine in Search.search_urls.keys():
            raise UnsupportedEngine

    @staticmethod
    def get_url(engine, query):
        return Search.search_urls[engine].format(query.replace(' ', '+'))

    @staticmethod
    def _b64decode_str(_str):
        return b64decode(_str.encode(Search.encoding)).decode(Search.encoding)

    @staticmethod
    def _b46encode_str(_str):
        return b64encode(_str.encode(Search.encoding)).decode(Search.encoding)

    def __init__(self, **kargs):
        self.engine = kargs.get('engine', 'google')
        Search.check_engine(self.engine)

        self.href_id = None
        self.query   = kargs.get('query', '')
        self.url     = Search.get_url(self.engine, self.query)
        self._set_base_url()

        self.session = requests.Session()
        self.session.headers.update(utils.generate_headers())
        self.timeout = kargs.get('timeout', 10)

        self.page      = ''
        self.index     = kargs.get('index', 0)
        self.save      = kargs.get('save', True)
        self.save_path = kargs.get('save_path', getenv('HOME', curdir))
        self._set_save_file()

    def _set_base_url(self):
        self.base_url = utils.get_base_url(Search.search_urls[self.engine])

    def _set_save_file(self):
        if S_ISREG(stat(self.save_path)[0]):
            self.save_file = self.save_path
            return

        self.save_file = re.match('(.+)' + sep + '?$', self.save_path).group(1) + sep + '.ima_cache'

    def set_query(self, query):
        self.query = query
        self.url   = Search.get_url(self.engine, self.query)
        self.index = 0

        return self

    def set_engine(self, engine):
        Search.check_engine(engine)

        self.index  = 0
        self.engine = engine
        self.url    = Search.get_url(self.engine, self.query)
        self._set_base_url()

        return self

    def set_url(self, url):
        self.url = url

        return self

    def _decode_url(self, url):
        return re.sub(r'%([a-fA-F0-9]{2})', lambda m: bytearray.fromhex(m.group(1)).decode(Search.encoding), url)

    def _extract_links(self):
        img_ext    = '\.' + '|'.join(utils.MIMETYPE_EXT.keys()) + '$'
        href_regex = {
            'google': [
                r'imgrefurl=[^&]+|(?:q|url|u)=https?://(?!(?:\w+\.)*google\.com)[^&]+',
                r'https://(?!(?:(?:\w+\.)*google\.com))',
            ],
            'yahoo': [
                r'https://r\.search\.yahoo\.com/.+/RO=\d+/RU=([^/]+)',
                r'https://(?!(?:(?:\w+\.)*yahoo\.(?:uservoice\.)?|(?:(?P<n>cc\.)|www\.)bing(?(n)j)\.)com)',
            ],
            'duckduckgo': [
                r'uddg=https?[^&]+',
                r'https://(?!(?:(?:\w+\.)*duckduckgo\.com))',
            ],
        }

        urls = set()
        dom  = BeautifulSoup(self.page, 'html.parser')
        for a in dom.find_all('a'):
            href = a.get('href')

            if href is None or not re.match('https?://', href):
                continue
            if self.engine == 'yahoo':
                matched = re.match(href_regex[self.engine][0], href)
                if matched:
                    url = self._decode_url(matched.group(1))
                    if re.match(href_regex[self.engine][1], url):
                        urls.add(url)
                continue

            if query := parse_url(href).query:
                matched = re.search(href_regex[self.engine][0], query) 
                if matched:
                    url = self._decode_url(matched.group().split('=')[1])
                    if not re.search(img_ext, url):
                        urls.add(url)

            elif re.match(href_regex[self.engine][1], href):
                urls.add(href)
        return urls

    def _get_request_data(self, sense):
        if self.engine == 'duckduckgo':
            return utils.get_post_data(self.page, '/html/', sense.capitalize())

        href_regex = {
            'google': r'/search\?q=[^&]+&.*(?<=&)start=(\d+)&',
            'yahoo': r'https://search\.yahoo\.com/search;[^?]+\?p=[^&]+&.*(?<=&)b=(\d+)&',
        }

        hrefs = utils.match_hrefs(self.page, href_regex[self.engine])
        if hrefs is None: return

        if len(hrefs) == 1:
            if sense == 'back':
                raise CannotGoBack
            return hrefs[0]['href']

        hrefs.sort(key = lambda m: m['id'])
        if len(hrefs) == 2:
            return hrefs[ 1 if sense == 'next' else 0 ]['href']

        #old          = self.href_id
        request_data = None
        for c in range(0, len(hrefs)):
            if not self.href_id:
                self.href_id = hrefs[c]['id']
                request_data = hrefs[c]['href']
                break

            #print("See: ", hrefs[c]['id'], ' and ', self.href_id)
            if self.href_id < hrefs[c]['id']:
                if sense == 'previous':
                    self.href_id = hrefs[c - 2]['id']
                    request_data = hrefs[c - 2]['href']
                else:
                    self.href_id = hrefs[c]['id']
                    request_data = hrefs[c]['href']
                break

            if sense == 'previous' and (
                self.href_id > hrefs[-1]['id'] or (
                    self.href_id == hrefs[-1]['id'] and c == (len(hrefs) - 1)
                )
            ):
                href         = hrefs[-2] if hrefs[-1]['id'] == self.href_id else hrefs[-1]
                self.href_id = href['id']
                request_data = href['href']
                break

        #print("id: ", old, "new id: ", self.href_id)
        #print("last: ", hrefs[-1])
        return request_data

    def _load_page(self, request_data):
        # Simple link to follow
        if isinstance(request_data, str):
            self.page = utils.http_x(
                'GET',
                self.session,
                utils.prepend_base_url(self.base_url, request_data),
                timeout = self.timeout
            ).text

        # HTTP POST, DuckDuckGO
        elif isinstance(request_data, dict):
            self.page = utils.http_x(
                'POST',
                self.session,
                utils.prepend_base_url(self.base_url, request_data['action']),
                data    = request_data['payload'],
                timeout = self.timeout
            ).text

    def _convert_links_to_image_objects(self, links):
        for link in links:
            yield Image(subject = self.query, url = link)

    def next(self, **kargs):
        save     = kargs.get('save', self.save)
        as_image = kargs.get('as_image', False)

        if self.index == 0:
            self.page = utils.http_x('GET', self.session, self.url, timeout = self.timeout).text
        else:
            request_data = self._get_request_data('next')
            if request_data is None:
                raise OutOfBoundError
            self._load_page(request_data)

        self.index += 1
        links = self._extract_links()
        if save:
            self._save(links)

        if as_image:
            return self._convert_links_to_image_objects(links)
        return list(links)

    def back(self, **kargs):
        save     = kargs.get('save', self.save)
        as_image = kargs.get('as_image', False)

        if self.index == 1 or self.index == 0:
            raise OutOfBoundError
        
        self._load_page(self._get_request_data('previous'))
        self.index -= 1

        links = self._extract_links()
        if save:
            self._save(links)

        if as_image:
            return self._convert_links_to_image_objects(links)
        return list(links)

    def get_links(self, n = 5, **kargs):
        if n < 1: raise OutOfBoundError

        links    = []
        save     = kargs.get('save', self.save)
        start    = kargs.get('start', True)
        as_image = kargs.get('as_image', False)

        if start is True:
            self.index = 1

        while len(links) < n:
            try:
                new_links = self.next()
                links += new_links
            except OutOfBoundError:
                break

        links = links[0:n]
        if save and len(links) > 0:
            self._save(links)

        if as_image is not None:
            return self._convert_links_to_image_objects(links)
        return links

    def _save(self, links):
        file       = self.save_file
        tmp_file   = file + '_tmp'
        links      = list(links)
        found_file = True

        makedirs(self.save_path, exist_ok = True)
        try:
            fd = open(file, 'r')
        except FileNotFoundError:
            found_file = False
            tmp_fd     = open(file, 'w')
        else:
            tmp_fd = open(tmp_file, 'w')

            while record := fd.readline():
                query, link, frequency = re.split(',', record)
                query, link = Search._b64decode_str(query), Search._b64decode_str(link)
                exists = False

                i = 0
                while i < len(links):
                    if link == links[i] and self.query == query:
                        frequency = str(int(frequency) + 1)
                        tmp_fd.write(re.match('(.+),', record).group(1) + ',' + frequency + "\n")
                        links.pop(i)
                        exists = True
                    i += 1
                if not exists:
                    tmp_fd.write(record)
        finally:
            for link in links:
                query = Search._b46encode_str(self.query)
                link  = Search._b46encode_str(link)
                tmp_fd.write('{0},{1},1\n'.format(query, link))

        if found_file:
            unlink(file)
            rename(tmp_file, file)

    def query_saves(self, **kargs):
        query      = kargs.get('query')
        query_like = kargs.get('query_like')
        frequency  = kargs.get('frequency')

        matched = []
        fd = open(self.save_file, 'r')
        while record := fd.readline():
            splited = list( map(Search._b64decode_str, re.split(',', record)) )

            if (
                 query and splited[0] == query
               ) or (
                 query_like and re.match(query_like, splited[0])
               ) or (
                 frequency and frequency == splited[2]
            ):
                matched += splited[1]
        return matched



class OutOfBoundError(Exception):
    """Cannot go back, currently at the first page"""

class HTTPResponseError(Exception):
    """Error returned by the server"""

class FileExistsError(Exception):
    """File already exists on filesystem"""

class UnsupportedEngine(Exception):
    """Search engine is unsupported"""

class CannotGoBack(Exception):
    """Cannot Go back, it isn't your fault"""



import sys
import optparse

from shutil   import get_terminal_size
from .version import __version__

def format_option(option):
    opts = []

    if option._short_opts:
        opts.append(option._short_opts[0])
    if option._long_opts:
        opts.append(option._long_opts[0])
    if len(opts) > 1:
        opts.insert(1, ', ')
    if option.takes_value():
        opts.append(' %s' % option.metavar)

    return ''.join(opts)

def ParseOptions():
    columns   = get_terminal_size().columns
    max_width = columns if columns else 80
    fmt       = optparse.IndentedHelpFormatter(width = max_width, max_help_position = 80)

    fmt.format_option_strings = format_option

    kwargs = {
        'version': __version__,
        'formatter': fmt,
        'usage': '%prog [OPTIONS] QUERY [QUERY...]',
        'conflict_handler': 'resolve',
    }

    parser = optparse.OptionParser(**kwargs)
    parser.add_option(
        '--version',
        action = 'version',
        help   = 'Print program version and exit.'
    )
    parser.add_option(
        '-h', '--help',
        action = 'help',
        help   = 'Print this help text and exit.'
    )
    parser.add_option(
        '-v', '--verbose',
        dest    = 'verbose',
        action  = 'store_true',
        default = False,
        help    = 'Print status messages to the standard output.'
    )
    parser.add_option(
        '-e', '--engine',
        dest    = 'engine',
        default = 'google',
        metavar = 'ENGINE',
        help    = 'Comma separated list of search engines to use, possible search engines are '
                  'duckduckgo, google, and yahoo. Engine defaults to "google". If more than '
                  'one search engine is specified, cycle through when connection fails too much.'
    )
    parser.add_option(
        '-n',
        dest    = 'n',
        type    = 'int',
        default = 4,
        metavar = 'NUM',
        help    = 'Print NUM results obtained from the `-s\' or `-i\' option or download NUM images '
                  'when neither of these options were specified.'
    )
    parser.add_option(
        '-u', '--number-sites',
        dest    = 'number_sites',
        type    = 'int',
        default = 10,
        metavar = 'NUM',
        help    = 'The number of websites to visit.'
    )
    parser.add_option(
        '-p', '--no-progress-bar',
        dest    = 'no_progress',
        action  = 'store_false',
        default = True,
        help    = 'Disable progress bar for downloads.'
    )
    parser.add_option(
        '-l', '--image-link',
        dest    = 'image_link',
        default = None,
        metavar = 'OUTPUT_FORMAT',
        help    = 'Output image links instead of downloading them. You can use the following specifiers to format '
                  'output: `{s}\' represents the score of the given image, `{d}\' its description, `{l}\' its '
                  'url, `{e}\' its file extension and `{w}\' the url from which the image link was extracted '
                  'e.g: we can for example format our output as such --> "image {l} has score {s}".'
    )
    parser.add_option(
        '-s', '--search-only',
        dest    = 'search',
        action  = 'store_true',
        default = False,
        help    = 'Query search engine and output search results only. This option is in conflict '
                  'with the `-l\' option.'
    )
    parser.add_option(
        '-m', '--image-count',
        dest    = 'image_count',
        type    = 'int',
        default = 2,
        metavar = 'NUM',
        help    = 'Set NUM as maximum number of image links to be extracted from a website.'
    )
    parser.add_option(
        '-d', '--dest-dir',
        dest    = 'dest_dir',
        type    = 'string',
        default = '.',
        metavar = 'DEST_DIR',
        help    = 'Specify the destination directory were downloaded files will be stored, '
                  'default to the current working directory of the executing program.'
    )
    parser.add_option(
        '-o', '--min-score',
        dest    = 'score',
        type    = 'int',
        default = 0,
        metavar = 'SCORE',
        help    = 'Define mininum score for images. This option should not be used with the `-s\' option.'
    )
    parser.add_option(
        '-r', '--retries',
        dest    = 'retries',
        type    = 'int',
        default = 0,
        metavar = 'NUM',
        help    = 'Number of retries before giving up if any connection fails.'
    )
    parser.add_option(
        '-x', '--retries-per-sites',
        dest    = 'retries_per_sites',
        type    = 'int',
        default = 0,
        metavar = 'NUM',
        help    = 'Number of retries per sites if any connection fails.'
    )
    parser.add_option(
        '-i', '--ignore-domains',
        dest    = 'no_domains',
        type    = 'string',
        default = None,
        metavar = 'DOMAIN',
        help    = 'A comma separated list of domains to ignore on search results.'
    )
    parser.add_option(
        '-w', '--overwrite',
        dest    = 'overwrite',
        action  = 'store_true',
        default = False,
        help    = 'Overwrite existing files, This option is in conflict with the `-a\' option.'
    )
    parser.add_option(
        '-a', '--auto-name',
        dest    = 'auto',
        action  = 'store_true',
        default = False,
        help    = 'Auto generate a new file name if a file name already exist in filesystem.'
    )
    parser.add_option(
        '-k', '--no-color',
        dest    = 'color',
        action  = 'store_false',
        default = True,
        help    = 'Disable ANSI colors'
    )
    parser.add_option(
        '-t', '--timeout',
        dest    = 'timeout',
        metavar = 'TIMEOUT',
        type    = 'int',
        default = 10,
        help    = 'Set connection timeout.'
    )
    parser.add_option(
        '-q', '--less-lines',
        dest    = 'more_lines',
        action  = 'store_true',
        default = False,
        help    = 'Wipe download progress bar after download has finished'
    )

    opts, args = parser.parse_args(sys.argv[1:])
    return opts, args



import re
import requests
import random

from requests.exceptions import InvalidURL

from bs4  import BeautifulSoup
from math import inf

from . import utils

encoding = utils.preferred_encoding()

class Image:

    @staticmethod
    def builtin_score(subject, content):
        if content is None or subject is None: return 0

        score   = 0
        content = content.casefold()
        for token in re.split(r'\s+', subject):
            token = token.casefold()
            if content.find(token.casefold()) != -1:
                score += 1
        return score

    def __init__(self, **kargs):
        self.url      = kargs.get('url', '')
        self.base_url = utils.get_base_url(self.url)
        self.session  = requests.Session()
        self.timeout  = kargs.get('timeout', 10)
        self.subject  = kargs.get('subject')

        self.session.headers.update(utils.generate_headers())

    def set_url(self, url):
        self.url      = url
        self.base_url = utils.get_base_url(url)


    def _get_link(self, tag_object, attributes, **kargs):
        score_link  = kargs.get('score_with', Image.builtin_score)
        min_score   = kargs.get('min_score', 0)
        use_content = kargs.get('use_content', True)

        content = tag_object.get('alt') if tag_object.string == None else tag_object.string.encode(encoding).decode(encoding)
        if use_content and content is None:
            return

        site_name = re.match(r'https?://(?:www\.)?([^.]+)', self.base_url).group(1)
        for attribute in attributes:
            url = tag_object.get(attribute)
            if not url:
                continue

            if attribute == 'srcset':
                # Just pick the first URL, NO OVERHEAD
                matched = re.search(r'\s*((?:https?:|/)?\S+(?<!,))\s*,?\s*(?:\d+(?:\.\d+)?(?:w|x))?', url)
                if not matched:
                    continue
                url = matched.group(1)
            else:
                if re.match('#|javascript:', url):
                    continue
                logo_regex = '/(?:[A-Za-z0-9]+[-_])*(?:logo|' + site_name + ')(?:[-_]?[A-Za-z0-9]+)*\\.' + '(:?' + '|'.join(utils.MIMETYPE_EXT.values()) + ')'
                if re.search(logo_regex, url):
                    continue

            if re.search(r'\.(?:gif|svg)$', url):
                continue
            if url.startswith('/'):
                url = utils.prepend_base_url(self.base_url, url)
            elif not ( url.startswith('data:image/') or url.startswith('http') ):
                continue

            if ( mime_type := utils.is_image(url, self.session, timeout = self.timeout) ) and (
                (score := score_link(self.subject, content)) >= min_score
            ):
                return {
                    'url': url,
                    'content': content if use_content else None,
                    'score': score,
                    'mime': mime_type,
                }

    def get_links(self, n = inf, **kargs):
        links     = []
        n         = inf if n is None else n
        self.page = utils.http_x('GET', self.session, self.url, timeout = self.timeout).text
        dom       = BeautifulSoup(self.page, 'html.parser')

        i     = 0
        links = []
        for tag_attributes in [
            [ 'img', [ 'data-src', 'src', 'srcset' ] ],
            '''[ 'a', [ 'href' ] ],'''
        ]:
            if i == n: break

            tag_objects = dom.find_all(tag_attributes[0])
            while i != n and len(tag_objects) > 0:
                j = random.choice( list(range(0, len(tag_objects))) )

                link = self._get_link(
                    tag_objects[j],
                    tag_attributes[1],
                    **kargs
                )
                if link and len( list(filter(lambda l: l == link, links)) ) == 0:
                    i += 1
                    links.append(link)
                    yield link

                tag_objects.pop(j)

    def download_from(self, link, **kargs):
        url, mime = None, None

        if isinstance(link, dict):
            url  = link.pop('url', None)
            mime = kargs.get('mime_type', None)
            if mime is None:
                mime = link.pop('mime')
        elif isinstance(link, str):
            url = link

        if url is None:
            raise requests.exceptions.InvalidURL
        return utils.download_image(url, self.session, mime_type = mime, **kargs) 
            
    def download(self, **kargs):
        n = kargs.pop('n', inf)

        for link in self.get_links(n):
            for stat in utils.download_image(
                link['url'],
                self.session,
                mime_type = link['mime'],
                **kargs
            ):
                yield stat



import pytest

from ima.search import Search

s = Search()

def next_back(s):
    r1 = s.next()
    r2 = s.next()
    r3 = s.next()
    b2 = s.back()
    b1 = s.back()

    assert r1 == b1
    assert r2 == b2

    f2 = s.next()
    assert f2 == r2

    f3 = s.next()
    assert r3 == f3

def test_next_back_google():
    s.set_engine('google')
    s.set_query('What is wikipedia')
    next_back(s)


def test_next_back_duckduckgo():
    s.set_engine('duckduckgo')
    s.set_query('What is wikipedia')
    next_back(s)

'''
test_object.test_next_back('duckduckgo', 'What is wikipedia')
test_object.test_next_back('yahoo', 'What is wikipedia')

test_object.test_next_back('google', 'Meaning of life')
test_object.test_next_back('duckduckgo', 'Meaning of love')
test_object.test_next_back('yahoo', 'How to be a rich man?')
'''



import sys
import random
import re

from os.path import basename

from requests.exceptions import (
    ConnectionError,
    ConnectTimeout,
    ReadTimeout,
    ChunkedEncodingError
)

from .        import exceptions as e
from .search  import Search
from .image   import Image
from .options import ParseOptions
from .utils   import (
    C,
    get_base_url,
    draw_bar,
    rewrite_text,
    hide_cursor,
    show_cursor,
    next_line,
    erase_up,
    MIMETYPE_EXT
)

ask = (
    'HIGH RESOLUTION IMAGES OF {0}',
    'IMAGES OF {0}',
    'I WANT IMAGES OF {0}',
    'DOWNLOAD IMAGES OF {0}',
    'PICTURES OF {0}'
    'FREE DOWNLOAD IMAGES OF {0}',
    'I WANT PICTURES OF {0}',
    'HIGH RESOLUTION PICTURES OF {0}',
)

def main():
    def _error(error):
        print(error, file = sys.stderr)

    me         = sys.argv[0]
    opts, args = ParseOptions()

    if len(args) == 0:
        _error("{0}: No query string(s), try `{0} -h' for more info.".format(me))
        exit(1)
    if opts.n < opts.image_count:
        _error("{0}: argument passed to `-n' must be > that of `-m'".format(me))
        exit(1)

    def _info(info, **kargs):
        print(info, file = sys.stdout, **kargs)

    tries = 0
    def _connection_handler(t = None, **kargs):
        nonlocal tries
        if kargs.get('nl'): next_line()
        t = (', ' + t) if t else ' to connect'
        if opts.retries == tries:
            _error('[{0}] Failed{1}.'.format(C('E', opts.color), t))
            show_cursor()
            exit(1)
        tries += 1
        _error('[{0}] Failed{1}.'.format(C('W', opts.color), t))

    def _interrupt_handler(**kargs):
        if kargs.get('nl'): next_line()
        _error('^Interrupted by user.')
        show_cursor()
        exit(1)

    def _in_domains(url):
        if opts.no_domains is None:
            return 0
        return len( list( filter(lambda d: url.endswith(d), re.split(',', opts.no_domains)) ) )

    hide_cursor()
    engine_index = 0
    engines      = re.split(',', opts.engine)
    search       = Search(engine = engines[0], timeout = opts.timeout, save = False)
    for query in args:
        search.set_query(random.choice(ask).format(query))

        n_failed = 0
        urls     = set()
        while True:
            try:
                urls = set(urls)
                while opts.number_sites != len(urls):
                    results = search.next()
                    urls    = urls.union(results[0:opts.number_sites - len(urls)])
                break
            except ConnectTimeout:
                _connection_handler('Connection timeout')
            except ReadTimeout:
                _connection_handler('Read timeout')
            except ConnectionError:
                _connection_handler()
            except e.HTTPResponseError:
                _error('[{0}] Could not search, HTTP response error.'.format(C('W', opts.color)))
                if len(engines) > 1 and n_failed == 5:
                    n_failed = 0
                    engine_index += 1
                    if len(engines) == engine_index:
                        engine_index = 0
                    search.set_engine(engines[engine_index])

                if opts.retries == tries:
                    if len(urls) > 0:
                        break
                    show_cursor()
                    exit(1)
                tries    += 1
                n_failed += 1
            except KeyboardInterrupt:
                _interrupt_handler()
            except e.OutOfBoundError:
                break

        urls = list(urls)
        random.shuffle(urls)

        n = 0
        for url in urls:
            if _in_domains(get_base_url(url)):
                continue

            if opts.search:
                _info(url if len(args) == 1 else '{0},{1}'.format(query, url))
                n += 1
                if opts.n == n:
                    show_cursor()
                    exit(0)
                continue

            ntries      = 0
            image_links = []
            image       = Image(url = url, subject = query, timeout = opts.timeout)

            if not opts.image_link:
                _info('[{0}] {1}'.format(C('S', opts.color), image.base_url))

            while True:
                try:
                    nb_lines = 0
                    if len(image_links) < opts.image_count:
                        for link in image.get_links(
                            opts.image_count,
                            min_score   = opts.score,
                            use_content = False if opts.score == 0 else True
                        ):
                            if image_links.count(link) != 0:
                                continue
                            if opts.verbose:
                                _info('[{0}] Found {1}'.format(C('I', opts.color), link))

                            if opts.image_link:
                                hash = { 'd': 'content', 'l': 'url', 's': 'score' }

                                def sub(m):
                                    if key := hash.get(m.group(1)):
                                        return link[key]
                                    if m.group(1) == 'e':
                                        return MIMETYPE_EXT[link['mime'].casefold()]
                                    if m.group(1) == 'w':
                                        return image.url
                                    _error("[{0}]: unrecognized format specifier `{1}'".format(C('E', opts.color), m.group(1)))
                                    exit(1)

                                _info(re.sub(r'(?<!\\)\{(e|w|l|s|d)\}', sub, opts.image_link))
                                image_links.append(link)
                                n += 1

                            while opts.image_link is None:
                                filename = None
                                try:
                                    if filename is not None:
                                        _info('[{0}] Retrying to download {1}'.format(C('I', opts.color), filename))

                                    # Fallback to default
                                    length, size = 0, ''
                                    for stat in image.download_from(
                                        link,
                                        path      = opts.dest_dir,
                                        overwrite = opts.overwrite,
                                        auto      = opts.auto
                                    ):
                                        if len(stat.keys()) == 1:
                                            read = stat['%']
                                            if size.startswith('0') or opts.no_progress is False:
                                                rewrite_text(read, length)
                                                length   = len(read)
                                                nb_lines = 1
                                            else:
                                                draw_bar(read.removesuffix('%'), 30)
                                                _info(read, end = '')
                                                nb_lines = 2
                                            continue

                                        filename = basename(stat['filename'])
                                        size     = stat['size']

                                        ln   = ' ' if size.startswith('0') or not opts.no_progress else '\n'
                                        _str = '[{0}]'.format(C('D', opts.color)) + ' Filename: {0}, Size: {1}'
                                        _info(_str.format(filename, size), end = ln)

                                    next_line()
                                    if nb_lines and opts.more_lines:
                                        erase_up(nb_lines + 1)

                                    n += 1
                                    image_links.append(link)
                                    break
                                except KeyboardInterrupt:
                                    _interrupt_handler(nl = True)
                                except ConnectTimeout:
                                    _connection_handler('Connection timeout', nl = True)
                                except ReadTimeout:
                                    _connection_handler('Read timeout', nl = True)
                                except ConnectionError:
                                    _connection_handler(nl = True)
                                    continue
                                except e.HTTPResponseError:
                                    next_line()
                                    _error('[{0}] HTTP error while downloading file {1}'.format(C('W', opts.color), filename))
                                    break
                            if opts.n == n:
                                show_cursor()
                                exit(1)
                            if len(image_links) == opts.image_count:
                                break
                    break
                except ChunkedEncodingError:
                    break
                except ConnectTimeout:
                    _connection_handler('Connection timeout')
                except ReadTimeout:
                    _connection_handler('Read timeout')
                except ConnectionError:
                    if ntries == opts.retries_per_sites:
                        break
                    ntries += 1
                    _connection_handler()
                except KeyboardInterrupt:
                    _interrupt_handler()
                except e.HTTPResponseError:
                    _error('[{0}] Could not follow {1} skipping ...'.format(C('E', opts.color), image.base_url))
                    break
    show_cursor()
    exit(0)
